/* FILE: output_defn.h                 -*-Mode: c++-*-
 *
 * Definitions for output classes, without implementation.  The
 * implementations are mainly in output.h (due to template
 * considerations).  This separate file exists so that Oxs_Ext
 * can include it to get the definition for its OxsExtUserOutput
 * class.  The Oxs_Output implementations need access to the
 * Oxs_Mesh and Oxs_MeshValue classes, which in turn need Oxs_Ext
 * definitions.  So if the Oxs_Output definitions and implementations
 * were in the same file, a circular dependency would arise.
 */


#ifndef _OXS_OUTPUT_DEFN
#define _OXS_OUTPUT_DEFN

#include <stdlib.h>

#include <string>

#include "meshvalue_defn.h"
#include "oc.h"
#include "oxsexcept.h"
#include "threevector.h"

OC_USE_STD_NAMESPACE;  // Specify std namespace, if supported
OC_USE_STRING;

/* End includes */

class Oxs_Director;   // Forward declarations
class Oxs_SimState;

class Oxs_Output {
private:
  Oxs_Director* director;  // Director instance this Oxs_Output
  /// instance is registered with.

  String owner_name;
  String output_name;
  String output_type;
  String output_units;

  BOOL cache_support; // If true, then caching is supported
  UINT4m cache_request_count;

  INT4m priority; // Output evaluation priority.  Smaller
  /// is higher priority.  Default priority is 0.  This
  /// can only be set from Oxs_Output::Register function,
  /// because it should remain constant throughout registration
  /// lifetime---the Oxsii and Boxsi scripts build an ordered
  /// output list by calling Oxs_Director::ListOutputObjects
  /// immediately after loading a problem, and use that list
  /// throughout the lifetime of the problem.

  // Disable copy constructor and assignment operator by declaring
  // them without defining them.
  Oxs_Output(const Oxs_Output&);
  Oxs_Output& operator=(const Oxs_Output&);

protected:
  BOOL ForceCacheIncrement(INT4m incr);
  /// Analogous to CacheRequestIncrement, but ignores
  /// cache_support value.  Returns 1 if increment was
  /// successful.  For use by child classes.

  BOOL GetMifOption(const char* label,String& value) const;

  const char* GetProblemName() const;

  void Setup(const char* owner_name_,
             const char* output_name_,
             const char* output_type_,
             const char* output_units_,
             BOOL cache_support_);
  // Called from public child Setup members, to initialize
  // private data.  Also resets cache_request_count to 0.

  Oxs_Director* GetDirector() const { return director; }

  void Register(Oxs_Director* director_,INT4m priority_);
  // Registers with director.  Automatic deregistration is set up
  // in ~Oxs_Output.

public:
  Oxs_Output();
  virtual ~Oxs_Output(); // Deregisters

  String OwnerName()  const { return owner_name;  }
  String OutputName() const { return output_name; }
  String LongName() const;
  String OutputType() const { return output_type; }
  String OutputUnits() const { return output_units; }
  virtual int SetOutputFormat(const String& format) =0;
  virtual String GetOutputFormat() const =0;

  virtual int Output(const Oxs_SimState*,Tcl_Interp*,int argc,
                     CONST84 char** argv) =0;

  // Cache flag control.  Returns 1 if increment was successful.
  // NB: The controlling cache_request_count variable gets reset
  //  each time Setup is called.  So CacheRequestIncrement should
  //  be called only _after_ Setup, and cache request needs must
  //  be re-requested any time Setup is re-called.
  BOOL CacheRequestIncrement(INT4m incr)
  {
    if(!cache_support) return 0; // Caching disabled
    return ForceCacheIncrement(incr);
  }

  UINT4m GetCacheRequestCount() const { return cache_request_count; }

  BOOL IsCacheSupported() const { return cache_support; }

  INT4m Priority() const { return priority; }

  // Output unit kludge added to support "user output" initialization
  // protocol.
  String GetUnits() const { return output_units; }
  void ChangeUnits(const String& new_units) {
    // Kludge added to support user output initialization protocol.
    output_units = new_units;
  }


};


//////////////////////////////////////////////////////////////////
// Templated cache class, used by templated output classes.
template<class T> struct Oxs_OutputCache {
public:
  Oxs_OutputCache() : state_id(0) {}
  UINT4m state_id;
  T value;
};

//////////////////////////////////////////////////////////////////
// Templated output classes.  The templated class T should be the
// embedding class.

////////////////////////
// Scalar output
template<class T> class Oxs_ScalarOutput:public Oxs_Output {
  typedef void (T::* GetScalar)(const Oxs_SimState&);
private:
  T* owner;
  GetScalar getscalar;
  /// Without the typedef, this would be
  ///    void (T::* getscalar)(const Oxs_SimState&)

  String output_format; // Default is "%.17g"

  // Disable copy constructor and assignment operator by declaring
  // them without defining them.
  Oxs_ScalarOutput(const Oxs_ScalarOutput<T> &);
  Oxs_ScalarOutput<T>& operator=(const Oxs_ScalarOutput<T>&);

public:
  Oxs_ScalarOutput() : Oxs_Output(),
    owner(NULL), getscalar(NULL), output_format("%.17g")
    {
      cache.value=0; // Safety
      CacheRequestIncrement(1);  // Temporary measure: until /**/
      /// cache management is implemented in Tcl interface,  /**/
      /// enable caching on all scalar outputs.              /**/
    }

  void Setup(T* owner_,
             const char* owner_name_,const char* output_name_,
             const char* output_units_,
             BOOL cache_support_,const GetScalar member_function)
  {
    owner = owner_;
    getscalar = member_function;
    Oxs_Output::Setup(owner_name_,output_name_,
                      "scalar",output_units_,
                      cache_support_);
  }

  void Register(Oxs_Director* director_,INT4m priority_)
  {
    Oxs_Output::Register(director_,priority_);
    if(GetDirector()!=NULL) {
      String value;
      if(GetMifOption("scalar_output_format",value)) {
        SetOutputFormat(value);
      }
    }
  }

  // Note: Base class member function GetMifOption() calls the director
  // to get the output format strings stored in the MIF options array.
  // We can't directly access member functions of director_ here because
  // this header file defines a forward declaration to the Oxs_Director
  // class, but does not #include "director.h".  We don't #include
  // "director.h" because director.h needs to include this file
  // (output.h) in order to get the definitions for the templated
  // Oxs_Output child classes Oxs_ScalarOutput and Oxs_VectorFieldOutput
  // (which it uses to provide total energy, total field and energy calc
  // counts).  Moreover, we can't move this constructor definition to
  // output.cc because of the template instantiation model being used
  // (what gcc calls the "Borland model"), which requires the inclusion
  // of full template definitions in each module using the template.
  // OTOH, the base class Oxs_Output is *not* templated, so
  // Oxs_Output::GetMifOption() *can* be located in output.cc, which
  // *can* #include "director.h", and so *can* call
  // director->GetMifOption().

  // Public data member for result drop-off/storage.
  // The owner should test GetCacheRequestCount() inside the output
  // calculation code, and adjust the cache as needed.
  Oxs_OutputCache<REAL8m> cache;

  virtual int SetOutputFormat(const String& format);
  virtual String GetOutputFormat() const { return output_format; }

  virtual int Output(const Oxs_SimState* state,Tcl_Interp* interp,
                     int argc,CONST84 char** argv);
  // The owner getscalar() function can assume on entry that the
  // cache value stored here is invalid, but it may become valid
  // during the lifetime of the getscalar() call (if getscalar()
  // calls some other function that sets the cache).
};

//////////////////////////////////////////////////////////////////
// Vector field output

// Define a non-templated cache class so clients can
// access stored values without having to know the owner.
class Oxs_VectorFieldOutputCache : public Oxs_Output {
private:
  // Disable copy constructor and assignment operator by declaring
  // them without defining them.
  Oxs_VectorFieldOutputCache(const Oxs_VectorFieldOutputCache&);
  Oxs_VectorFieldOutputCache&
     operator=(const Oxs_VectorFieldOutputCache&);

public:
  Oxs_VectorFieldOutputCache() : Oxs_Output() {}

  // Public data member for result drop-off/storage.  The owner should
  // test GetCacheRequestCount() inside the output calculation code, and
  // adjust the cache as needed.
  Oxs_OutputCache<Oxs_MeshValue<ThreeVector> > cache;

  virtual BOOL UpdateCache(const Oxs_SimState*) = 0;
  /// Returns 1 on success, 0 if caching not supported.
  /// Throws an exception if an error occurs.
};

// Main vector field output class
template<class T> class Oxs_VectorFieldOutput
  :public Oxs_VectorFieldOutputCache {
  typedef void (T::* GetVectorField)(const Oxs_SimState&);
private:
  T* owner;
  GetVectorField getvectorfield;
  /// Without the typedef, this would be
  ///    void (T::* getvectorfield)(const Oxs_SimState&)

  BOOL output_format_writeheaders; // Default is 1

  String output_format_meshtype;   // Default is "rectangular"
  String output_format_datatype;   // Default is "binary"
  String output_format_precision;  // Default is "8"

  // Disable copy constructor and assignment operator by declaring
  // them without defining them.
  Oxs_VectorFieldOutput(const Oxs_VectorFieldOutput<T>&);
  Oxs_VectorFieldOutput<T>& operator=(const Oxs_VectorFieldOutput<T>&);

public:

  Oxs_VectorFieldOutput() : Oxs_VectorFieldOutputCache(),
    owner(NULL), getvectorfield(NULL),
    output_format_writeheaders(1),
    output_format_meshtype("rectangular"),
    output_format_datatype("binary"),
    output_format_precision("8")
    {}

  void Setup(T* owner_,
             const char* owner_name_,const char* output_name_,
             const char* output_units_,
             BOOL cache_support_,
             const GetVectorField member_function)
    {
      owner = owner_;
      getvectorfield = member_function;
      Oxs_Output::Setup(owner_name_,output_name_,
                        "vector field",output_units_,
                        cache_support_);
    }

  void Register(Oxs_Director* director_,INT4m priority_)
  {
    Oxs_Output::Register(director_,priority_);
    if(GetDirector()!=NULL) {
      String value;
      if(GetMifOption("vector_field_output_format",value)) {
        SetOutputFormat(value);
      }
      if(GetMifOption("vector_field_output_meshtype",value)) {
        SetOutputMeshtype(value);
      }
      if(GetMifOption("vector_field_output_writeheaders",value)) {
        SetOutputWriteheaders(atoi(value.c_str()));
      }
    }
  }

  // Note: Base class member function GetMifOption() calls the director
  // to get the output format strings stored in the MIF options array.
  // We can't directly access member functions of director_ here because
  // this header file defines a forward declaration to the Oxs_Director
  // class, but does not #include "director.h".  We don't #include
  // "director.h" because director.h needs to include this file
  // (output.h) in order to get the definitions for the templated
  // Oxs_Output child classes Oxs_ScalarOutput and Oxs_VectorFieldOutput
  // (which it uses to provide total energy, total field and energy calc
  // counts).  Moreover, we can't move this constructor definition to
  // output.cc because of the template instantiation model being used
  // (what gcc calls the "Borland model"), which requires the inclusion
  // of full template definitions in each module using the template.
  // OTOH, the base class Oxs_Output is *not* templated, so
  // Oxs_Output::GetMifOption() *can* be located in output.cc, which
  // *can* #include "director.h", and so *can* call
  // director->GetMifOption().

  virtual int SetOutputFormat(const String& format);
  virtual String GetOutputFormat() const;

  int SetOutputWriteheaders(BOOL writeheaders) {
    output_format_writeheaders = writeheaders;
    return 0;
  }
  BOOL GetOutputWriteheaders() const {
    return output_format_writeheaders;
  }

  int SetOutputMeshtype(const String& meshtype);
  String GetOutputMeshtype() const;

  virtual BOOL UpdateCache(const Oxs_SimState*);
  /// Returns 1 on success, 0 if caching not supported.
  /// Throws an exception if an error occurs.

  virtual int Output(const Oxs_SimState* state,Tcl_Interp* interp,
                     int argc,CONST84 char** argv);
};

#endif // _OXS_OUTPUT_DEFN
