/* FILE: randomvectorfield.cc      -*-Mode: c++-*-
 *
 * Random vector field object, derived from Oxs_VectorField class.
 *
 */

#include <string>

#include "oc.h"
#include "nb.h"

#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "randomvectorfield.h"
#include "energy.h"             // Needed to make MSVC++ 5 happy

OC_USE_STRING;

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_RandomVectorField);

/* End includes */


// Constructor
Oxs_RandomVectorField::Oxs_RandomVectorField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_VectorField(name,newdtr,argstr),
    min_norm(-1.),max_norm(-1.),mincubed(-1.),maxcubed(-1.)
{
  // Process arguments.
  min_norm=GetRealInitValue("min_norm");
  max_norm=GetRealInitValue("max_norm");

  VerifyAllInitArgsUsed();

  if(min_norm>max_norm) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
                "Requested min_norm value (%g) bigger than max_norm (%g)",
                static_cast<double>(min_norm),
                static_cast<double>(max_norm));
    throw Oxs_Ext::Error(this,buf);
  }

  // Cube weights are necessary to get a uniform sampling by volume.
  // Pre-compute for convenience.  (Assume no overflow.)
  mincubed = min_norm*min_norm*min_norm;
  maxcubed = max_norm*max_norm*max_norm;
}


void
Oxs_RandomVectorField::Value
(const ThreeVector& /* pt */,
 ThreeVector& value) const
{
  if(min_norm == max_norm) {
    // No spread in magnitudes
    value.Random(max_norm);
  } else {
    // Vary both direction and magnitude.
    // Magnitude needs to be adjusted on a cube scale
    // to get uniform sampling by volume.
    REAL8m randval = Oc_UnifRand();
    REAL8m mag = pow((1-randval)*mincubed+randval*maxcubed,1./3.);
    value.Random(mag);
  }
}

void
Oxs_RandomVectorField::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<ThreeVector>& array) const
{
  array.AdjustSize(mesh);
  UINT4m i,size=mesh->Size();
  if(min_norm == max_norm) {
    // No spread in magnitudes
    for(i=0;i<size;i++) {
      array[i].Random(max_norm);
    }
  } else {
    // Vary both direction and magnitude.
    // Magnitude needs to be adjusted on a cube scale
    // to get uniform sampling by volume.
    for(i=0;i<size;i++) {
      REAL8m randval = Oc_UnifRand();
      REAL8m mag = pow((1-randval)*mincubed+randval*maxcubed,1./3.);
      array[i].Random(mag);
    }
  }
}

void
Oxs_RandomVectorField::FillMeshValueOrthogonal
(const Oxs_Mesh* mesh,
 const Oxs_MeshValue<ThreeVector>& orth_array,
 Oxs_MeshValue<ThreeVector>& array) const
{
  if(!orth_array.CheckMesh(mesh)) {
    String msg=
      String("Mesh size incompatibility detected in"
             " Oxs_RandomVectorField::FillMeshValueOrthogonal()"
             " method of instance ")
      + String(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Magnitude needs to be adjusted on a square scale
  // to get uniform sampling by volume.
  REAL8m minsq = min_norm*min_norm;
  REAL8m maxsq = max_norm*max_norm;
  BOOL no_spread = 0;
  if(min_norm == max_norm) no_spread = 1;

  REAL8m dot=0.0;
  ThreeVector vec,tempvec;
  UINT4m i,size=mesh->Size();
  array.AdjustSize(mesh);
  for(i=0;i<size;i++) {
    // First produce vector perpendicular to orth_array[i]
    tempvec=orth_array[i];
    tempvec.SetMag(1.0);
    int j=0;
    while(1) {
      vec.Random(1.0);
      dot = tempvec*vec;
      if(fabs(dot)<0.99) break;
      if(++j>32) {
        String msg=
          String("Programming error?  Unable to obtain vector"
                 " in specified plane in"
                 " Oxs_RandomVectorField::FillMeshValueOrthogonal()"
                 " method of instance ")
          + String(InstanceName());
        throw Oxs_Ext::Error(msg.c_str());
      }
    }
    vec -= dot*tempvec;

    // Adjust to proper size
    REAL8m mag = max_norm;
    if(!no_spread) {
      // Adjust to square scale
      REAL8m randval = Oc_UnifRand();
      mag = sqrt((1-randval)*minsq+randval*maxsq);
    }
    vec.SetMag(mag);
    array[i] = vec;
  }
}
