/* FILE: chanwrap.h                 -*-Mode: c++-*-
 *
 * C++ wrappers for Tcl channels.  These wrappers are exception safe,
 * and also provide low-overhead, buffered byte-at-a-time access to
 * the channels.  Support is only for strictly read-only or write-only
 * channels.  Also, once a channel is attached to one of these objects,
 * only access through the object is allowed; e.g., direct calls to
 * Tcl_Read will almost certainly result in undetected reordering or
 * lossage of bytes on the input channel.  If the channel supports
 * random seeks (via Tcl_Tell and Tcl_Seek), then the channel can
 * be detached from the object, after which all the standard Tcl
 * I/O calls can be used again.
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2007-03-20 01:05:39 $
 * Last modified by: $Author: donahue $
 */

#ifndef _NB_CHANWRAP
#define _NB_CHANWRAP

#include <tcl.h>

#include "oc.h"

/* End includes */     /* Optional directive to build.tcl */

////////////////////////////////////////////////////////////////////////
// Wrapper for read-access Tcl channels
class Nb_InputChannel {
public:
  Nb_InputChannel();

  ~Nb_InputChannel() { Close(); }

  void OpenFile(const char* filename,const char* translation);

  void AttachChannel(const Tcl_Channel& newchan,const char* description) {
    Close();
    chan = newchan;
    file_descriptor.Dup(description);
  }

  Tcl_Channel DetachChannel() {
    if(chan==NULL) return NULL;
    if(Tcl_Seek(chan,buf_offset - buf_fillsize,SEEK_CUR)<0) {
      OC_THROW(Oc_Exception
               (__FILE__,__LINE__,"Nb_InputChannel","DetachChannel",
                1000+2000,
                "Detach failed due to invalid seek"
                " on file \"%.2000s\".",
                file_descriptor.GetStr()));
    }
    Tcl_Channel chancopy = chan;
    chan=NULL;
    buf_fillsize = buf_offset = 0;
    file_descriptor.Dup("");
    return chancopy;
  }

  int Close() {
    int errcode = TCL_OK;
    if(chan!=NULL) { errcode = Tcl_Close(NULL,chan); }
    chan=NULL;
    buf_fillsize = buf_offset = 0;
    file_descriptor.Dup("");
    return errcode;
  }

  int ReadChar() {
    if(buf_offset<buf_fillsize) {
      return static_cast<int>
        (static_cast<unsigned char>(buf[buf_offset++]));
    }
    if((buf_fillsize = FillBuf())>0) {
      buf_offset=1;
      return static_cast<int>(static_cast<unsigned char>(buf[0]));
    }
    return -1;
  }

private:
  Oc_AutoBuf file_descriptor;
  Tcl_Channel chan;
  int buf_fillsize;
  int buf_offset;
  Oc_AutoBuf buf;

  int FillBuf();

  // Disable copy constructor and assignment operator
  Nb_InputChannel(const Nb_InputChannel&);
  Nb_InputChannel& operator=(const Nb_InputChannel&);
};

////////////////////////////////////////////////////////////////////////
// Wrapper for write-access Tcl channels
//
// WARNING: THIS CODE IS COMPLETELY UNUSED AND UNTESTED!
//                                      17-Mar-2007, mjd
//
#define NB_OUTPUTCHANNEL_BUFSIZE 65536
class Nb_OutputChannel {
public:
  Nb_OutputChannel();

  ~Nb_OutputChannel() { Close(); }

  void OpenFile(const char* filename,const char* translation);

  void AttachChannel(const Tcl_Channel& newchan,const char* description) {
    Close();
    chan = newchan;
    file_descriptor.Dup(description);
  }

  void DetachChannel() {
    if(chan==NULL)return;
    WriteBuf();
    chan=NULL;
    buf_offset = 0;
    file_descriptor.Dup("");
  }

  int Close() {
    int errcode = TCL_OK;
    if(chan!=NULL) {
      WriteBuf();
      errcode = Tcl_Close(NULL,chan);
    }
    chan=NULL;
    buf_offset = 0;
    file_descriptor.Dup("");
    return errcode;
  }

  int WriteChar(int byte) {
    char ch = static_cast<char>(byte);
    buf[buf_offset] = ch; // Assumes buf_offset < NB_OUTPUT_CHANNEL_BUFSIZE
    if( (++buf_offset) < NB_OUTPUTCHANNEL_BUFSIZE ) {
      return static_cast<int>(static_cast<unsigned char>(ch));
    }
    int checkcode = WriteBuf();
    buf_offset=0; // Safety
    if(checkcode<0) return checkcode;
    return static_cast<int>(static_cast<unsigned char>(ch));
  }

private:
  Oc_AutoBuf file_descriptor;
  Tcl_Channel chan;
  int buf_offset;
  Oc_AutoBuf buf;

  int WriteBuf();

  // Disable copy constructor and assignment operator
  Nb_OutputChannel(const Nb_OutputChannel&);
  Nb_OutputChannel& operator=(const Nb_OutputChannel&);
};

#endif // _NB_CHANWRAP
