/* FILE: imgobj.cc                 -*-Mode: c++-*-
 *
 * C++ wrapper for 2D image data.  At present, the data are stored as
 * 24-bit color values, red+green+blue, in consecutive elements of a
 * char array.  The interface is intended to be flexible enough that,
 * potentially, the backend could be changed to support monochromatic
 * storage (as single elements of char array, to reduce memory
 * footprint) or wider data types (to provide greater color depth).
 *
 * This class is designed with primary emphasis on minimizing memory
 * use for large images.  Access speed is only a secondary criterion.
 *
 * NOTE: This class does not require Tk.  However, if the input file
 * is not in format supported by this code, then avf2ppm is called,
 * and avf2ppm does use Tk.
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2008-06-27 06:52:59 $
 * Last modified by: $Author: donahue $
 */

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "oc.h"
#include "imgobj.h"

/* End includes */     /* Optional directive to build.tcl */

#define NB_IMGOBJ_ERRBUFSIZE 4096  // Buffer size used for error messages
#define NB_IMGOBJ_CMDBUFSIZE 16384 // Buffer size used to contruct and
                                  /// hold Tcl commands

////////////////////////////////////////////////////////////////////////

// Read one text value routine, for PPM files.  This routine is needed
// by both text and binary PPM formats, because the header in both
// cases is in ASCII text.
INT4m Nb_ImgObj::ReadPPMTextInt(Nb_InputChannel& chan,INT4m& value)
{ //    Returns 0 on success, or on error returns
  //         1 => EOF
  //         2 => Input contains character not allowed in an integer value

  int byte;

  // Read past any initial whitespace or comments on chan
  while((byte=chan.ReadChar())>=0) {
    if(isspace(byte)) continue;
    if(byte=='#') {
      // Comment; read until end of line
      while((byte=chan.ReadChar())>=0) {
        if(byte=='\n' || byte=='\r') break;
      }
    } else {
      break;
    }
  }

  if(byte<0) return 1; // EOF
  if(byte<'0' || '9'<byte) return 2; // Illegal value

  // Otherwise, byte holds the first non-whitespace,
  // non-comment character, i.e., the first character
  // of the token.
  INT4m tmpval = byte - '0';

  // Read until first non-numeric value, shifting values into tmpval;
  while((byte=chan.ReadChar())>=0) {
    byte -= '0';
    if(byte<0 || byte>9) break; // Non-numeric value
    tmpval = 10*tmpval+byte;
  }

  value = tmpval;
  return 0;
}

INT4m Nb_ImgObj::ReadThreeIntPPMText
(Nb_InputChannel& chan,
 INT4m& value1,
 INT4m& value2,
 INT4m& value3)
{
  INT4m err;
  if((err=ReadPPMTextInt(chan,value1))!=0) {
    return err;
  }
  if((err=ReadPPMTextInt(chan,value2))!=0) {
    return err;
  }
  return ReadPPMTextInt(chan,value3);
}

INT4m Nb_ImgObj::ReadThreeIntPPMBin8
(Nb_InputChannel& chan,
 INT4m& value1,
 INT4m& value2,
 INT4m& value3)
{
  if((value1 = static_cast<INT4m>(chan.ReadChar()))<0 ||
     (value2 = static_cast<INT4m>(chan.ReadChar()))<0 ||
     (value3 = static_cast<INT4m>(chan.ReadChar()))<0) {
    return 1; // Presumed EOF
  }
  return 0;
}

INT4m Nb_ImgObj::ReadThreeIntPPMBin16
(Nb_InputChannel& chan,
 INT4m& value1,
 INT4m& value2,
 INT4m& value3)
{
  int tmp1,tmp2;
  if((tmp1 = chan.ReadChar())<0 || (tmp2 = chan.ReadChar())<0) {
    return 1; // Presumed EOF
  }
  value1 = tmp1*256 + tmp2;
  if((tmp1 = chan.ReadChar())<0 || (tmp2 = chan.ReadChar())<0) {
    return 1; // Presumed EOF
  }
  value2 = tmp1*256 + tmp2;
  if((tmp1 = chan.ReadChar())<0 || (tmp2 = chan.ReadChar())<0) {
    return 1; // Presumed EOF
  }
  value3 = tmp1*256 + tmp2;
  return 0;
}


void Nb_ImgObj::LoadFile(const char* infilename)
{
  Free();

  filename.Dup(infilename);

  Nb_InputChannel chan;
  chan.OpenFile(filename,"binary");

  // See if we file is a Microsoft .bmp format that we can read.
  {
    Tcl_Channel tclchan = chan.DetachChannel();
    if(ReadMSBitmap(tclchan)) return;  // Yes.

    // No. Reset channel.
    Tcl_Seek(tclchan,0,SEEK_SET);
    chan.AttachChannel(tclchan,filename);
  }

  // Check file type.  NOTE: PPM headers are in ASCII format.  The
  // following magic number checks should be modified to compare
  // against pure values (e.g., 'P'=0x50, '3'=0x33) if this code is
  // ever ported to an EBCDIC. :^)
  ImgFmt image_type = IMG_UNKNOWN;
  int byte1 = chan.ReadChar();
  int byte2 = chan.ReadChar();
  int byte3 = chan.ReadChar();
  if(byte1=='P' && byte2 == '3' && isspace(byte3)) {
    image_type = P3;
  } else if(byte1=='P' && byte2 == '6' && isspace(byte3)) {
    image_type = P6_8; // Temporary assignment.  This may be
    // reassigned to P6_16 depending on the maxvalue value.
  }
  // Add BMP support here if an when it becomes available.
  if(image_type == IMG_UNKNOWN) {
    // Unsupported format.  Close and send to any2ppm for conversion.
    // Re-purpose chan to point to command channel.
    char buf[NB_IMGOBJ_CMDBUFSIZE];
    if(chan.Close()!=TCL_OK) {
      // Close error.
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                         NB_IMGOBJ_ERRBUFSIZE+2000,
                         "Error closing input channel for file \"%.2000s\".",
                         filename.GetStr()));
    }

    Tcl_Interp* interp=Oc_GlobalInterpreter();
    Tcl_SavedResult saved;
    Tcl_SaveResult(interp, &saved);

    // Find name of null device on platform
    Oc_Snprintf(buf,sizeof(buf),
                "[Oc_Config RunPlatform] GetValue path_device_null");
    if(Tcl_Eval(interp,buf)!=TCL_OK) {
      Tcl_RestoreResult(interp, &saved);
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                         NB_IMGOBJ_ERRBUFSIZE,
                         "Can't determine null device name."));
    }
    Oc_AutoBuf nulldevice = Tcl_GetStringResult(interp);
    Tcl_ResetResult(interp);

    // Construct command line for launching any2ppm
    Oc_Snprintf(buf,sizeof(buf),"Oc_Application CommandLine any2ppm"
                " -noinfo -format P6 -o - {%s}",
                filename.GetStr());
    if(Tcl_Eval(interp,buf)!=TCL_OK) {
      Tcl_RestoreResult(interp, &saved);
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                         NB_IMGOBJ_ERRBUFSIZE+2000,
                         "Can't find any2ppm program to"
                         " convert image file \"%.2000s\" to PPM P6 format.",
                         filename.GetStr()));
    }
    Oc_Snprintf(buf,sizeof(buf),"%s < %s 2> %s",Tcl_GetStringResult(interp),
                nulldevice.GetStr(),nulldevice.GetStr());
    Tcl_ResetResult(interp);

    // Parse command line
    int argc;
    CONST84 char ** argv;
    if(Tcl_SplitList(NULL,buf,&argc,&argv)!=TCL_OK) {
      Tcl_RestoreResult(interp, &saved);
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                         NB_IMGOBJ_ERRBUFSIZE+2000,
                         "Can't parse command string: \"%.2000s\"",buf));
    }

    // Launch any2ppm command pipe
    Tcl_Channel tmpchan = Tcl_OpenCommandChannel(interp,argc,argv,TCL_STDOUT);
    if(tmpchan==NULL) {
      Oc_Exception errmsg(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                 NB_IMGOBJ_ERRBUFSIZE+strlen(Tcl_GetStringResult(interp)),
                 "any2ppm launch error: %s",Tcl_GetStringResult(interp));
      Tcl_RestoreResult(interp, &saved);
      Tcl_Free((char *)argv);
      OC_THROW(errmsg);
    }
    Tcl_Free((char *)argv);
    Tcl_RestoreResult(interp, &saved);

    if(Tcl_SetChannelOption(NULL,tmpchan,OC_CONST84_CHAR("-translation"),
			    OC_CONST84_CHAR("binary"))!=TCL_OK ||
       Tcl_SetChannelOption(NULL,tmpchan,OC_CONST84_CHAR("-buffering"),
			    OC_CONST84_CHAR("full"))!=TCL_OK ||
       Tcl_SetChannelOption(NULL,tmpchan,OC_CONST84_CHAR("-buffersize"),
			    OC_CONST84_CHAR("65536"))!=TCL_OK) {
      // Channel configuration error.
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                NB_IMGOBJ_ERRBUFSIZE+2000,
                "Unable to configure avf2ppm command pipe"
                " for file \"%.2000s\".",
                filename.GetStr()));
    }
    Oc_Snprintf(buf,sizeof(buf),"avf2ppm < \"%s\" |",filename.GetStr());
    chan.AttachChannel(tmpchan,buf);

    // Read new file type info
    byte1 = chan.ReadChar();
    byte2 = chan.ReadChar();
    byte3 = chan.ReadChar();
    if(byte1=='P' && byte2 == '3' && isspace(byte3)) {
      image_type = P3;
    } else if(byte1=='P' && byte2 == '6' && isspace(byte3)) {
      image_type = P6_8; // Temporary assignment.  This may be
      // reassigned to P6_16 depending on the maxvalue value.
    }
    if(image_type == IMG_UNKNOWN) {
      // Not a PPM file?
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                NB_IMGOBJ_ERRBUFSIZE+2000,
                "Failure in PPM conversion for file \"%.2000s\".",
                filename.GetStr()));
    }
  }

  // Read PPM header
  INT4m errcode = 0;
  errcode += ReadPPMTextInt(chan,width);
  errcode += ReadPPMTextInt(chan,height);
  errcode += ReadPPMTextInt(chan,maxvalue);
  if(errcode!=0) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Read error in header for file \"%.2000s\".",
             filename.GetStr()));
  }

  if(width<1 || height<1 || maxvalue<0) {
    // Bad file data.
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Bad header reading file \"%.2000s\".",
             filename.GetStr()));
  }
  data.SetSize(height,NB_IMGOBJ_DATUMSIZE*width);

  if(image_type == P6_8 && maxvalue>255) {
    // Correction.  This is actually P6_16 format
    image_type = P6_16;
    if(maxvalue>65535) {
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
               NB_IMGOBJ_ERRBUFSIZE+2000,
               "Error in Nb_ImgObj::LoadFile: "
               "Bad header reading file \"%.2000s\"; PPM P6 file "
               "has maxvalue=%d>65535.",filename.GetStr(),maxvalue));
    }
  }

  // Point ReadThreeValues member function pointer to the
  // appropriate routine for the given image type.
  INT4m (*ReadThreeValues)(Nb_InputChannel&,INT4m&,INT4m&,INT4m&)
    = NULL;
  switch(image_type) {
  case P3:
    ReadThreeValues = &Nb_ImgObj::ReadThreeIntPPMText;
    break;
  case P6_8:
    ReadThreeValues = &Nb_ImgObj::ReadThreeIntPPMBin8;
    break;
  case P6_16:
    ReadThreeValues = &Nb_ImgObj::ReadThreeIntPPMBin16;
    break;
  default:
    {
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
               NB_IMGOBJ_ERRBUFSIZE+2000,
               "Bad format reading file \"%.2000s\"; Programming error?",
               filename.GetStr()));
    }
  }

  assert(NB_IMGOBJ_DATUMSIZE==3);
  INT4m red,green,blue;
  if(maxvalue<=255) {
    // Usual case: Input data fit into one byte each
    for(INT4m i=0;i<height;++i) {
      for(INT4m j=0;j<NB_IMGOBJ_DATUMSIZE*width;j+=NB_IMGOBJ_DATUMSIZE) {
        if((errcode = (*ReadThreeValues)(chan,red,green,blue))!=0) {
          OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                   NB_IMGOBJ_ERRBUFSIZE+2000,
                   "Bad data or EOF in file \"%.2000s\"; "
                   "offset=(%d,%d), errcode=%d.",
                   filename.GetStr(),static_cast<int>(i),
                   static_cast<int>(j/NB_IMGOBJ_DATUMSIZE),
                   static_cast<int>(errcode)));
        }
        data[i][j]   = static_cast<unsigned char>(red);
        data[i][j+1] = static_cast<unsigned char>(green);
        data[i][j+2] = static_cast<unsigned char>(blue);
      }
    }
  } else { // maxvalue>255
    // Nb_ImgObj data storage only supports 1 byte widths.  If
    // maxvalue is larger than 255, adjust for this on import.
    REAL8m scale = 255.5/static_cast<REAL8m>(maxvalue);
    maxvalue = 255;
    for(INT4m i=0;i<height;++i) {
      for(INT4m j=0;j<NB_IMGOBJ_DATUMSIZE*width;j+=NB_IMGOBJ_DATUMSIZE) {
        if((errcode = (*ReadThreeValues)(chan,red,green,blue))!=0) {
          OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
                   NB_IMGOBJ_ERRBUFSIZE+2000,
                   "Bad data or EOF in file \"%.2000s\"; "
                   "offset=(%d,%d), errcode=%d.",
                   filename.GetStr(),static_cast<int>(i),
                   static_cast<int>(j/NB_IMGOBJ_DATUMSIZE),
                   static_cast<int>(errcode)));
        }
        data[i][j]   =
          static_cast<unsigned char>(floor(static_cast<REAL8m>(red)*scale));
        data[i][j+1] =
          static_cast<unsigned char>(floor(static_cast<REAL8m>(green)*scale));
        data[i][j+2] =
          static_cast<unsigned char>(floor(static_cast<REAL8m>(blue)*scale));
      }
    }
  }

  if(chan.Close()!=TCL_OK) {
    // Close error.
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","LoadFile",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Error closing input channel for file \"%.2000s\".",
             filename.GetStr()));
  }

  return;
}

////////////////////////////////////////////////////////////////////////
// MsBitmap and related classes
struct Nb_ImgObj_RGBQuad {
public:
  // Note: MS RGBQuad's are ordered "Blue Green Red Reserved"
  BYTE Blue;
  BYTE Green;
  BYTE Red;
  BYTE Reserved;
};

struct Nb_ImgObj_MSHeader {
public:
  BYTE  Type[2];
  UINT4 FileSize;
  UINT2 Reserved1;
  UINT2 Reserved2;
  UINT4 OffBits;
  UINT4 BmiSize;
  UINT4 Width;
  UINT4 Height;
  UINT2 Planes;
  UINT2 BitCount;
  UINT4 Compression;
  UINT4 SizeImage;
  UINT4 XPelsPerMeter;
  UINT4 YPelsPerMeter;
  UINT4 ClrUsed;
  UINT4 ClrImportant;
  Nb_ImgObj_RGBQuad* Palette;

  void FreePalette() {
    if(Palette!=NULL) delete[] Palette;
    Palette = NULL;
  }

  void AllocPalette(int size) {
    FreePalette();
    Palette=new Nb_ImgObj_RGBQuad[size];
  }

  Nb_ImgObj_MSHeader() :
    FileSize(0),Reserved1(0),Reserved2(0),OffBits(0),
    BmiSize(0),Width(0),Height(0),Planes(0),BitCount(0),Compression(0),
    SizeImage(0),XPelsPerMeter(0),YPelsPerMeter(0),
    ClrUsed(0),ClrImportant(0),Palette(NULL),
    PaletteSize(0)
  {
    Type[0]=Type[1]='\0';
  }

  ~Nb_ImgObj_MSHeader() { FreePalette(); }

  int FillHeader(Tcl_Channel chan);
  ///   Returns 1 if successful, 0 if the input does not appear
  /// to be a valid MS bitmap.  Adjusts for alignment and byte
  /// ordering as needed.

  UINT4m GetFileRowSize() {
    return 4*((Width*BitCount+31)/32); // Rows lie on a 4-byte bdry.
  }

  UINT4m GetPaletteSize() {
    return PaletteSize;
  }
private:
  UINT4m PaletteSize;
};

int Nb_ImgObj_MSHeader::FillHeader(Tcl_Channel chan)
{ // Reads members one at a time to avoid differing packing
  // alignment restrictions on various machines.  Returns 1
  // on success.

  // Read and check signature
  Tcl_Read(chan,(char *)Type,2*sizeof(BYTE));
  if(Type[0]!='B' || Type[1]!='M') return 0;

  // Read rest of header
  Tcl_Read(chan,(char *)&FileSize,sizeof(UINT4));
  Tcl_Read(chan,(char *)&Reserved1,sizeof(UINT2));
  Tcl_Read(chan,(char *)&Reserved2,sizeof(UINT2));
  Tcl_Read(chan,(char *)&OffBits,sizeof(UINT4));

  // MS Bitmap format is little-endian ordered;
  // Adjust byte ordering, if necessary.
#if BYTEORDER != 4321
#if BYTEORDER != 1234
#error "Unsupported byte-order format"
#endif
  Oc_Flip4(&FileSize);
  Oc_Flip2(&Reserved1); Oc_Flip2(&Reserved2);
  Oc_Flip4(&OffBits);
#endif

  // Check file size.  Is there no Tcl C interface to the 'file size'
  // proc???
  Oc_SeekPos bmistart=Tcl_Tell(chan);
  Oc_SeekPos real_size=Tcl_Seek(chan,0,SEEK_END);
  Tcl_Seek(chan,bmistart,SEEK_SET);
  if(real_size != static_cast<Oc_SeekPos>(-1) &&
     real_size != static_cast<Oc_SeekPos>(FileSize)) return 0;

  // Otherwise, read in rest of header
  Tcl_Read(chan,(char *)&BmiSize,sizeof(UINT4));
  Tcl_Read(chan,(char *)&Width,sizeof(UINT4));
  Tcl_Read(chan,(char *)&Height,sizeof(UINT4));
  Tcl_Read(chan,(char *)&Planes,sizeof(UINT2));
  Tcl_Read(chan,(char *)&BitCount,sizeof(UINT2));
  Tcl_Read(chan,(char *)&Compression,sizeof(UINT4));
  Tcl_Read(chan,(char *)&SizeImage,sizeof(UINT4));
  Tcl_Read(chan,(char *)&XPelsPerMeter,sizeof(UINT4));
  Tcl_Read(chan,(char *)&YPelsPerMeter,sizeof(UINT4));
  Tcl_Read(chan,(char *)&ClrUsed,sizeof(UINT4));
  Tcl_Read(chan,(char *)&ClrImportant,sizeof(UINT4));

  // MS Bitmap format is little-endian ordered;
  // Adjust byte ordering, if necessary.
#if BYTEORDER != 4321
#if BYTEORDER != 1234
#error "Unsupported byte-order format"
#endif
  Oc_Flip4(&BmiSize);
  Oc_Flip4(&Width);          Oc_Flip4(&Height);
  Oc_Flip2(&Planes);         Oc_Flip2(&BitCount);
  Oc_Flip4(&Compression);    Oc_Flip4(&SizeImage);
  Oc_Flip4(&XPelsPerMeter);  Oc_Flip4(&YPelsPerMeter);
  Oc_Flip4(&ClrUsed);        Oc_Flip4(&ClrImportant);
#endif

  // Fill palette
  FreePalette();
  Tcl_Seek(chan,bmistart+BmiSize,SEEK_SET); // Position at
  /// end of BITMAPINFOHEADER

  PaletteSize = ClrUsed;
  if(PaletteSize==0) {
    switch(BitCount)
      {
      case  1: PaletteSize=2;    break;
      case  4: PaletteSize=16;   break;
      case  8: PaletteSize=256;  break;
      case 24: PaletteSize=0;    break;  // True color
      default: return 0;     // Invalid BitCount value
      }
  }
  if(PaletteSize>0) {
    AllocPalette(PaletteSize);
    Tcl_Read(chan,(char *)Palette,PaletteSize*sizeof(Palette[0]));
  }

  return 1;
}

void
Nb_ImgObj::ReadBmp1bit
(Tcl_Channel chan,
 Nb_ImgObj_MSHeader* header)
{
  // Note: Rows in the input file run from bottom to top.
  if(header->GetPaletteSize() != 2) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp1bit",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Invalid palette size (%u) specified in header---palette"
             " size should be 2; Bad header reading file \"%.2000s\".",
             header->GetPaletteSize(),filename.GetStr()));
  }
  Nb_ImgObj_RGBQuad offbit = header->Palette[0];
  Nb_ImgObj_RGBQuad onbit  = header->Palette[1];
  const int filerowsize = header->GetFileRowSize();
  if((width-1)/8 >= filerowsize) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp1bit",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Image width (%d) is wider than file row width (%d)"
             " as specified in header;"
             " Bad header reading file \"%.2000s\".",
             width,8*filerowsize,filename.GetStr()));
  }
  Oc_AutoBuf buf;
  buf.SetLength(filerowsize);
  unsigned char* const chbuf = (unsigned char*)buf.GetStr();
  for(INT4m i=height-1;i>=0;--i) {
    if(Tcl_Read(chan,(char*)chbuf,filerowsize) != filerowsize) {
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp1bit",
              NB_IMGOBJ_ERRBUFSIZE+2000,
              "Premature end-of-file reading image file \"%.2000s\".",
              filename.GetStr()));
    }
    INT4m j=0;
    while(j<width) {
      unsigned int datum=chbuf[j/8];
      for(unsigned int mask=0x80; mask!=0 && j<width ; mask>>=1, ++j) {
        if((datum & mask)==0) {
          data[i][NB_IMGOBJ_DATUMSIZE*j]   = offbit.Red;
          data[i][NB_IMGOBJ_DATUMSIZE*j+1] = offbit.Green;
          data[i][NB_IMGOBJ_DATUMSIZE*j+2] = offbit.Blue;
        } else {
          data[i][NB_IMGOBJ_DATUMSIZE*j]   = onbit.Red;
          data[i][NB_IMGOBJ_DATUMSIZE*j+1] = onbit.Green;
          data[i][NB_IMGOBJ_DATUMSIZE*j+2] = onbit.Blue;
        }
      }
    }
  }
}

void
Nb_ImgObj::ReadBmp4bit
(Tcl_Channel chan,
 Nb_ImgObj_MSHeader* header)
{
  // Note: Rows in the input file run from bottom to top.
  const Nb_ImgObj_RGBQuad* pal = header->Palette;
  const int filerowsize = header->GetFileRowSize();
  if( (width-1)/2 >= filerowsize) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp4bit",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Image width (%d) is wider than file row width (%d)"
             " as specified in header;"
             " Bad header reading file \"%.2000s\".",
             width,2*filerowsize,filename.GetStr()));
  }
  const UINT4m palsize = header->GetPaletteSize();
  Oc_AutoBuf buf;
  buf.SetLength(filerowsize);
  unsigned char* const chbuf = (unsigned char*)buf.GetStr();
  for(INT4m i=height-1;i>=0;--i) {
    if(Tcl_Read(chan,(char*)chbuf,filerowsize) != filerowsize) {
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp4bit",
               NB_IMGOBJ_ERRBUFSIZE+2000,
               " Premature end-of-file reading image file \"%.2000s\".",
               filename.GetStr()));
    }
    INT4m j=0;
    while(j<width) {
      unsigned int datum=chbuf[j/2];
      UINT4 index1=  datum & 0x0F;        // Note bit-order!
      UINT4 index0= (datum & 0xF0)>>4;
      if(index0 >= palsize) {
        // palsize will usually be 16, in which case this check isn't
        // needed.  If this check turns out to be a performance
        // problem, then we should probably handle the palsize<16
        // case in a separate branch.
        OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp4bit",
                 NB_IMGOBJ_ERRBUFSIZE+2000,
                 "Palette index request (%u) is outside palette range"
                 " (0-%u); Bad data reading file \"%.2000s\".",
                 index0,palsize,filename.GetStr()));
      }
      data[i][NB_IMGOBJ_DATUMSIZE*j]   = pal[index0].Red;
      data[i][NB_IMGOBJ_DATUMSIZE*j+1] = pal[index0].Green;
      data[i][NB_IMGOBJ_DATUMSIZE*j+2] = pal[index0].Blue;
      if(++j>=width) break;
      if(index1 >= palsize) {
        OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp4bit",
                 NB_IMGOBJ_ERRBUFSIZE+2000,
                 "Palette index request (%u) is outside palette range"
                 " (0-%u); Bad data reading file \"%.2000s\".",
                 index1,palsize,filename.GetStr()));
      }
      data[i][NB_IMGOBJ_DATUMSIZE*j]   = pal[index1].Red;
      data[i][NB_IMGOBJ_DATUMSIZE*j+1] = pal[index1].Green;
      data[i][NB_IMGOBJ_DATUMSIZE*j+2] = pal[index1].Blue;
      ++j;
    }
  }
}

void
Nb_ImgObj::ReadBmp8bit
(Tcl_Channel chan,
 Nb_ImgObj_MSHeader* header)
{
  // Note: Rows in the input file run from bottom to top.
  const Nb_ImgObj_RGBQuad* pal = header->Palette;
  const int filerowsize = header->GetFileRowSize();
  if(width > filerowsize) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp8bit",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Image width (%d) is wider than file row width (%d)"
             " as specified in header;"
             " Bad header reading file \"%.2000s\".",
             width,filerowsize,filename.GetStr()));
  }
  const UINT4m palsize = header->GetPaletteSize();
  Oc_AutoBuf buf;
  buf.SetLength(filerowsize);
  unsigned char* const chbuf = (unsigned char*)buf.GetStr();
  for(INT4m i=height-1;i>=0;--i) {
    if(Tcl_Read(chan,(char*)chbuf,filerowsize) != filerowsize) {
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp8bit",
               NB_IMGOBJ_ERRBUFSIZE+2000,
               "Premature end-of-file reading image file \"%.2000s\".",
               filename.GetStr()));
    }
    for(INT4m j=0;j<width;++j) {
      UINT4m index = chbuf[j];
      if(index >= palsize) {
        // palsize will usually be 256, in which case this check
        // isn't needed.  If this check turns out to be a performance
        // problem, then we should probably handle the palsize<256
        // case in a separate branch.
        OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp8bit",
                 NB_IMGOBJ_ERRBUFSIZE+2000,
                 "Palette index request (%u) is outside palette range"
                 " (0-%u); Bad data reading file \"%.2000s\".",
                 index,palsize,filename.GetStr()));
      }
      data[i][NB_IMGOBJ_DATUMSIZE*j]   = pal[index].Red;
      data[i][NB_IMGOBJ_DATUMSIZE*j+1] = pal[index].Green;
      data[i][NB_IMGOBJ_DATUMSIZE*j+2] = pal[index].Blue;
    }
  }
}

void
Nb_ImgObj::ReadBmp24bit
(Tcl_Channel chan,
 Nb_ImgObj_MSHeader* header)
{ // True color input file; 24 bits per pixel, no palette.
  // Note 1: Byte order in input file is "Blue Green Red".
  // Note 2: Rows end on 4-byte boundary.
  // Note 3: Rows in the input file run from bottom to top.
  const int filerowsize = header->GetFileRowSize();
  if(width > filerowsize) {
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp24bit",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Image width (%d) is wider than file row width (%d)"
             " as specified in header;"
             " Bad header reading file \"%.2000s\".",
             width,filerowsize,filename.GetStr()));
  }
  Oc_AutoBuf buf;
  buf.SetLength(filerowsize);
  unsigned char* const chbuf = (unsigned char*)buf.GetStr();
  for(INT4m i=height-1;i>=0;--i) {
    if(Tcl_Read(chan,(char*)chbuf,filerowsize) != filerowsize) {
      OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadBmp24bit",
               NB_IMGOBJ_ERRBUFSIZE+2000,
               "Premature end-of-file reading image file \"%.2000s\".",
               filename.GetStr()));
    }
    for(INT4m j=0;j<NB_IMGOBJ_DATUMSIZE*width;j+=NB_IMGOBJ_DATUMSIZE) {
      data[i][j]   = chbuf[j+2];  // Red
      data[i][j+1] = chbuf[j+1];  // Green
      data[i][j+2] = chbuf[j];    // Blue
    }
  }
}

int
Nb_ImgObj::ReadMSBitmap(Tcl_Channel  chan)
{ // Attempts to fill in data, one row at a time.  Returns
  // 1 on success, 0 if the data on chan is either not in
  // MS .bmp format, or is in an MS .bmp format we don't
  // support.

  Nb_ImgObj_MSHeader header;

  if(header.FillHeader(chan) == 0) return 0;

  // Looks like an MS bitmap file.  Is it in a subformat
  // we support?
  if(header.Planes!=1         ||      // Only known value
     header.Compression!=0    ||      // No compression
     (header.BitCount!=1 && header.BitCount!=4 && header.BitCount!=8
      && header.BitCount!=24) // Bits per pixel check
     ) {
    return 0;
  }

  width = header.Width;
  height = header.Height;
  maxvalue = 255;  // I guess...

  data.SetSize(height,NB_IMGOBJ_DATUMSIZE*width);

  // Fill in data, using routine determined by bits-per-pixel value.
  switch(header.BitCount) {
  case 1:
    ReadBmp1bit(chan,&header);
    break;
  case 4:
    ReadBmp4bit(chan,&header);
    break;
  case 8:
    ReadBmp8bit(chan,&header);
    break;
  case 24:
    ReadBmp24bit(chan,&header);
    break;
  default:
    // The subformat check above should insure that we don't get here.
    OC_THROW(Oc_Exception(__FILE__,__LINE__,"Nb_ImgObj","ReadMSBitmap",
             NB_IMGOBJ_ERRBUFSIZE+2000,
             "Bad format reading file \"%.2000s\"; Programming error?",
             filename.GetStr()));
  }

  return 1;
}

